@@ -251,7 +251,7 @@ void ffInitBatteryOptions(FFBatteryOptions* options)
         ffPrintBatteryHelpFormat,
         ffGenerateBatteryJsonConfig
     );
-    ffOptionInitModuleArg(&options->moduleArgs);
+    ffOptionInitModuleArg(&options->moduleArgs, "");
     options->temp = false;
     options->tempConfig = (FFColorRangeConfig) { 60, 80 };
     options->percent = (FFColorRangeConfig) { 50, 20 };
