@@ -6,7 +6,7 @@
 
 #include <math.h>
 
-#define FF_DISPLAY_NUM_FORMAT_ARGS 16
+#define FF_DISPLAY_NUM_FORMAT_ARGS 15
 
 static int sortByNameAsc(FFDisplayResult* a, FFDisplayResult* b)
 {
@@ -156,7 +156,6 @@ void ffPrintDisplay(FFDisplayOptions* options)
                 {FF_FORMAT_ARG_TYPE_DOUBLE, &ppi, "ppi"},
                 {FF_FORMAT_ARG_TYPE_UINT8, &result->bitDepth, "bit-depth"},
                 {FF_FORMAT_ARG_TYPE_BOOL, &result->hdrEnabled, "hdr-enabled"},
-                {FF_FORMAT_ARG_TYPE_BOOL, &result->wcgEnabled, "wcg-enabled"},
             }));
         }
     }
@@ -326,7 +325,6 @@ void ffGenerateDisplayJsonResult(FF_MAYBE_UNUSED FFDisplayOptions* options, yyjs
         yyjson_mut_obj_add_uint(doc, obj, "rotation", item->rotation);
         yyjson_mut_obj_add_uint(doc, obj, "bitDepth", item->bitDepth);
         yyjson_mut_obj_add_bool(doc, obj, "hdrEnabled", item->hdrEnabled);
-        yyjson_mut_obj_add_bool(doc, obj, "wcgEnabled", item->wcgEnabled);
 
         switch (item->type)
         {
@@ -361,7 +359,6 @@ void ffPrintDisplayHelpFormat(void)
         "Pixels per inch (PPI) - ppi",
         "Bits per color channel - bit-depth",
         "True if high dynamic range (HDR) is enabled - hdr-enabled",
-        "True if wide color gamut (WCG) is enabled - wcg-enabled",
     }));
 }
 
